"""Module for the translations."""

__all__: list[str] = ["load", "get"]

# Built-ins
from typing import Any
from pathlib import Path
import json
import os

# Third-party
import yaml

# Anint
from .utils import parse_key
from .exceptions import TranslationError, MultipleSameLocaleError

# Storage dictionary.
data: dict[str, Any] = {}


def load_yaml(filepath: str) -> dict[str, Any]:
    """Parse the given YAML file and return a dictionary.

    :param filepath: Absolute path to the YAML file.
    :return: Dictionary of the parsed YAML file.
    """
    with open(filepath, "r", encoding="utf8") as file:
        return yaml.safe_load(file)


def load_json(filepath: str) -> dict[str, Any]:
    """Parse the given JSON file and return a dictionary.

    :param filepath: Absolute path to the JSON file.
    :return: Dictionary of the parsed JSON file.
    """
    with open(filepath, "r", encoding="utf8") as file:
        return json.load(file)


def load(path_to_locale_directory: str) -> None:
    """Load the translation from the given path_to_locale_directory.

    :param str path_to_locale_directory: Path to the translation file or directory containing the translation file.
    :return: None.
    :raise FileNotFoundError: If the requested file does not exist.
    """
    for filepath in os.listdir(path_to_locale_directory):
        locale, extension = os.path.splitext(filepath)
        path_to_locale: Path = Path(path_to_locale_directory, filepath)
        extension = extension[1:]
        if locale not in data:
            match extension:
                case "yaml":
                    data[locale] = load_yaml(str(path_to_locale))
                case "json":
                    data[locale] = load_json(str(path_to_locale))
        else:
            raise MultipleSameLocaleError(locale)


def get(locale: str, key: str) -> str:
    """Parse the locale data as is for the specified key.

    :param str locale: Specify which locale to get the translation from.
    :param str key: A string of dict keys combined by dots.
    :return: The translation for the current specified locale.
    :raise TranslationError: If the key raises a KeyError or if the referred value is not of type str.
    """
    try:
        parsed_keys: list[str] = parse_key(key)
        value: dict = data[locale]
        for parsed_key in parsed_keys:
            value = value[parsed_key]

        if isinstance(value, str):
            return value
        else:
            raise TranslationError(
                "{key} argument does not represent a localizable value".format(key=key)
            )
    except KeyError:
        raise TranslationError(key)



# Built-ins
import os
import pathlib

# Anint
from anint import translations
from anint import exceptions as err

# Third-party
import pytest

path0: pathlib.Path = pathlib.Path(os.path.dirname(__file__)) / "locales" / "yaml"
translations.load(str(path0))


class TestTranslations:
    def test_return_value0(self):
        assert translations.get("en", "greetings.hello") == "Hello"

    def test_return_value1(self):
        assert translations.get("jp", "greetings.hello") == "こんにちは"

    def test_exception0(self):
        with pytest.raises(err.TranslationError):
            translations.get("en", "greetings.bye")



"""Models for Anint."""

# Built-ins
from typing import Optional

# Anint
from .exceptions import TranslationError
from . import translations


class Translator:
    """Translator class."""

    def __init__(
        self,
        locales: list[str],
        locale: Optional[str] = None,
        fallbacks: Optional[list[str]] = None,
    ) -> None:
        """Initialize Translator class object.

        :param locales: List of available locales.
        :param locale: Specified locale.
        :param fallbacks: Fallback locales.
        :return: None.
        """
        self.locales: list[str] = locales
        self.locale: Optional[str] = locale
        self.fallbacks: Optional[list[str]] = fallbacks

    def set_locale(self, locale: str) -> None:
        """Change the locale setting to the specified locale.

        :param str locale: The desired language code.
        :return: None.
        :raise ValueError: If locale not in the list of available locales.
        """
        if locale in self.locales:
            self.locale = locale
        else:
            raise ValueError(locale)

    def translate(self, key: str, *args) -> str:
        """Returns the translation for the specified key.

        :param str key: A string sequence of dict keys connected by dots.
        :param args: Passed onto the translation to be formatted if there are any placeholders.
        :return: The translation for the currently specified language setting.
        """
        if not self.locale:
            raise ValueError(
                "Not specified locale exists. Set a locale to localize in."
            )

        try:
            value: str = translations.get(self.locale, key)
        except TranslationError:
            if not self.fallbacks:
                raise TranslationError(key)

            for fallback in self.fallbacks:
                try:
                    value = translations.get(fallback, key)
                    break
                except TranslationError:
                    pass
            else:
                raise TranslationError(key)

        return value.format(*args)



"""Constants of Anint."""

# Built-ins
from enum import StrEnum


class AnintDict(StrEnum):
    """Container enum for the project and its configuration."""

    TOOL = "tool"
    ANINT = "anint"
    LOCALES = "locales"
    LOCALE = "locale"
    FALLBACKS = "fallbacks"
    PATH = "path"



"""Loaded data from the Anint configuration file if any."""

__all__: list[str] = ["data", "instance_data"]

# Built-ins
import configparser, tomllib, os
from configparser import ConfigParser
from typing import Any, Optional

# Anint
from .utils import get_file_extension, csv_to_list
from .exceptions import AnintConfigError
from .constants import AnintDict


def fetch_config_file() -> Optional[str]:
    """Return the absolute path to the highest priority Anint configuration file.

    :return: Absolute path to the highest priority Anint configuration file.
    :raise AnintConfigError: If no matching config file is found.
    """
    for filepath in [
        "{anint}.ini".format(anint=AnintDict.ANINT),
        ".{anint}.ini".format(anint=AnintDict.ANINT),
        "pyproject.toml",
        "{anint}.cfg".format(anint=AnintDict.ANINT),
    ]:
        if os.path.exists(filepath):
            return os.path.abspath(filepath)
    else:
        return None


def load_raw_ini(filepath: str) -> dict[str, Any]:
    """Return a dictionary of the Anint configuration from any .ini or .cfg file.

    :param str filepath: Path to the .ini or .cfg file.
    :return: Dictionary of the loaded values as is. *Beware that comma seperated values will not be converted to a list of elements.
    """
    config_data: ConfigParser = configparser.ConfigParser()
    config_data.read(filepath)
    try:
        return dict(config_data.items(AnintDict.ANINT))
    except configparser.NoSectionError:
        return {}


def load_raw_toml(filepath: str) -> dict[str, Any]:
    """Return a dictionary of the Anint configuration from the pyproject.toml file.

    :param str filepath: Path to the .toml file.
    :return: Dictionary of the loaded values as is. *Beware that comma seperated values will not be converted to a list of elements.
    """
    try:
        with open(filepath, "rb") as file:
            return tomllib.load(file)[AnintDict.TOOL][AnintDict.ANINT]
    except KeyError:
        return {}


def load_config() -> dict[str, Any]:
    """Return a dictionary of the Anint configuration.

    | An INI configuration file:
    | [anint]
    | locales = mn, en, jp
    | locale = jp
    | fallbacks = en
    | path = ./tests/locales
    |
    | Will be loaded as:
    | {
    | locales: ["mn", "en", "jp"],
    | locale: "jp",
    | fallbacks: ["en"],
    | path: "${PATH_TO_WORKING_DIRECTORY}/tests/locales"
    | }

    :return: Dictionary of the loaded Anint configuration. *Expects all keys to have a corresponding value, will assign empty values otherwise.
    :raise AnintConfigError: If the loaded file is not a .ini or .cfg or .toml file.
    """
    config_data: dict[str, Any] = {}
    if filepath := fetch_config_file():
        filename: str = os.path.basename(filepath)
        extension: str = get_file_extension(filename)
        if extension == "ini" or extension == "cfg":
            config_data = load_raw_ini(filepath)
        elif extension == "toml":
            config_data = load_raw_toml(filepath)
        else:
            raise AnintConfigError(
                "The following {filename} is not a recognized configuration file format.".format(
                    filename=filename
                )
            )

    # Normalize loaded config data.
    config_data[AnintDict.LOCALES] = csv_to_list(config_data.get(AnintDict.LOCALES, ""))
    config_data[AnintDict.LOCALE] = config_data.get(AnintDict.LOCALE, "")
    config_data[AnintDict.FALLBACKS] = csv_to_list(
        config_data.get(AnintDict.FALLBACKS, "")
    )
    config_data[AnintDict.PATH] = os.path.abspath(config_data.get(AnintDict.PATH, ""))

    return config_data


data: dict[str, Any] = load_config()
instance_data: dict[str, Any] = {
    key: value
    for key, value in data.items()
    if key in [AnintDict.LOCALES, AnintDict.LOCALE, AnintDict.FALLBACKS]
}
